* Display Extensions V2.01, last update 21/02/03
* QLiberator directive $$asmb=flp1_display_cde,0,10
*
* Brief description and syntax of basic extensions:
*
* ADDRESS returns base address for given screen channel
* LET screen_base_address = ADDRESS(#channel)
*
* BYTES returns number of bytes per line for given window channel
* LET display_width = BYTES(#channel)
*
* DMODE returns current screen display mode (0 or 8, for example)
* LET screen_mode = DMODE(#channel)
*
* SYS_VAR returns base address of system variables area
* LET variables = SYS_VAR
*
* The FLIM_n functions use iop.flim trap to find window limits. For
* primary channels, this is screen size, for secondaries this is the
* outline limits
*
* FLIM_X and FLIM_Y return origin of largest window channel
* LET x = FLIM_X(#channel)
* LET y = FLIM_Y(#channel)
* FLIM_W and FLIM_H return limits of width and height respectively
* LET wide = FLIM_W(#channel)
* LET high = FLIM_H(#channel)

* MOVEMEM shifts a given number of bytes between two addresses, always
* from the first address to the second address, so overwriting (e.g. to
* clear an area of memory) is possible
* MOVEMEM from_address,to_address,bytes

* The new extensions in v2.00 are as follows
* (assistance from Marcel Kilgus gratefully acknowledged):
* LET v$  = OS_VER$            returns operating system version n.nn
* LET v$  = PTRVER$(#channel)  returns pointer version n.nn
* LET env = PTR_ENV(#channel)  returns 0 if no PE, 1 if PE present
* LET wm  = WIN_MAN(#channel)  returns 0 if no WMAN, 1 if WMAN present
* LET gd  = GD2(#channel)      returns 0 if no GD2, 1 if GD2 present

* Added in v2.01 with information from Rich Mellor:
* LET v$  = WMAVER$(#channel)  returns Window Manager version number

* link into basic as extensions
        lea     exts,a1         ;point to list of extensions
        move.w  $110,a2         ;bp.init
        jmp     (a2)            ;link extensions

* list of extension names and definitions
exts    dc.w    1               ;just one procedure
        dc.w    movemem-*       ;location
        dc.b    7,'MOVEMEM'     ;name
        dc.w    0               ;end of procedures list
        
        dc.w    14              ;14 functions
        dc.w    address-*       ;location of ADDRESS
        dc.b    7,'ADDRESS'     ;name of function
        dc.w    bytes-*         ;location of BYTES
        dc.b    5,'BYTES'       ;name of function
        dc.w    dmode-*         ;location of DMODE
        dc.b    5,'DMODE'       ;name of function
        dc.w    sys_var-*       ;location of SYS_VAR
        dc.b    7,'SYS_VAR'     ;name of function
        dc.w    flim_x-*        ;location of FLIM_X
        dc.b    6,'FLIM_X',0    ;name of function
        dc.w    flim_y-*        ;location of FLIM_Y
        dc.b    6,'FLIM_Y',0    ;name of function
        dc.w    flim_w-*        ;location of FLIM_W
        dc.b    6,'FLIM_W',0    ;name of function
        dc.w    flim_h-*        ;location of FLIM_H
        dc.b    6,'FLIM_H',0    ;name of function
        dc.w    ptr_env-*       ;location of PTR_ENV
        dc.b    7,'PTR_ENV'     ;name of function
        dc.w    win_man-*       ;location of WIN_MAN
        dc.b    7,'WIN_MAN'     ;name of function
        dc.w    gd2-*           ;location of GD2
        dc.b    3,'GD2'         ;name of function
        dc.w    ptr_ver-*       ;location of PTRVER$
        dc.b    7,'PTRVER$'     ;name of function
        dc.w    os-*            ;location of OS_VER$
        dc.b    7,'OS_VER$'     ;name of function
        dc.w    wma_ver-*       ;location of WMAVER$
        dc.b    7,'WMAVER$'     ;anme of function
        dc.w    0               ;end of function definitions list
        
* code for the MOVEMEM procedure
* MOVEMEM from_address,to_address,bytes
movemem move.w  $118,a2         ;ca.gtlin to fetch long integers
        jsr     (a2)            ;try to fetch 3 long integers
        bne     return          ;oops
        subq.w  #3,d3           ;3 parameters?
        bne     badparam        ;oops
*check parameter values that addresses and bytes are positive
        tst.l   0(a6,a1.l)      ;source address
        blt     badparam        ;invalid negative address?
        move.l  0(a6,a1.l),a2   ;source address into a2
        tst.l   4(a6,a1.l)      ;destination address
        blt     badparam        ;invalid negative address
        move.l  4(a6,a1.l),a3   ;destination address into a3
        tst.l   8(a6,a1.l)      ;number of bytes
        beq.s   mvedone         ;zero bytes, don't do anything
        blt     badparam        ;invalid negative number of bytes
        move.l  8(a6,a1.l),d1   ;number of bytes to be moved
movelop move.b  (a2)+,(a3)+     ;shift a byte
        subq.l  #1,d1           ;count the bytes
        bne.s   movelop         ;keep going until all done
mvedone moveq   #0,d0           ;no error report
        rts                     ;return to basic

* LET adr = ADDRESS(#channel)
* returns base address for given screen channel
address move.w  $118,a2         ;ca.gtlin
        jsr     (a2)            ;fetch 1 parameter
        bne     return          ;oops...
        subq.w  #1,d3           ;1 parameter?
        bne     badparam        ;oops...
* fine, we have one parameter on the stack, so for any
* further errors, we can return a suitable negative number
* instead of an error report
* all the functions use d7 to indicate how many bytes more are
* required on top of any parameters already on the RI stack (the
* channel number is a 4 byte value, so the norm is 2 extra bytes
* needed to return a 6 byte float value)
        moveq   #2,d7           ;2 extra bytes will be needed
        move.l  0(a6,a1.l),d0   ;get channel number to d0
        bmi     retfp1          ;return the negative d0 value
        bsr     chan2id         ;convert d0 channel# to ID in a0
        bmi     retfp1          ;oops, return the negative number

        moveq   #9,d0           ;prepare to use the channel
        
        moveq   #$32,d1         ;sd.scrb offset within definition block
        moveq   #-1,d3          ;infinite timout
        lea.l   extop_l,a2      ;the routine to be invoked by sd.extop
        trap    #3              ;use sd.extop to call the routine
* d1 now contains the value required (screen base address)
        move.l  d1,d0           ;put required value in d0 for retfp
        bra     retfp1          ;and return it

* this routine is invoked from sd.extop to fetch a value indexed by d1
* from the base of the definition block in a0
* (Thank you Norman Dunbar for originally showing me how to use
* sd.extop)
extop_l       move.l    0(a0,d1.w),d1   ;fetch long value required
              moveq     #0,d0           ;ensure no error
              rts                       ;and go back to caller
extop_w       move.w    0(a0,d1.w),d1   ;fetch word value required
              moveq     #0,d0           ;ensure no error
              rts                       ;and go back to caller

* LET bytes_per_line = BYTES(#channel)
* returns display width for given screen channel
* AH/JM roms (QDOS 1.03 or earlier) will return wrong result
* so return fixed value of 128 for these roms
bytes   move.w  $118,a2         ;ca.gtlin
        jsr     (a2)            ;fetch 1 parameter
        bne     return          ;oops
        subq.w  #1,d3           ;1 parameter?
        bne     badparam        ;oops
* fine, we have one parameter on the stack, so for any further errors
* we can return a suitable negative number instead of an error report

* all the functions use d7 to indicate how many bytes more are required
* on top of any parameters already on the RI stack (the channel number
* is a 4 byte value, so the norm is 2 extra bytes needed to return a 
* 6 byte float value)
        moveq   #2,d7           ;2 extra bytes wil be needed
        move.l  0(a6,a1.l),d0   ;get channel number into d0
        bmi     retfp1          ;return the negative d0 value
        bsr     chan2id         ;convert channel# in d0 to ID in a0
        bmi     retfp1          ;oops, return the negative number
* at this point:
* A0=channel ID   A1=RI ptr    d7=extra RI stack bytes wanted

* Thanks to Ralf Rekoendt for this tip originally
* test if AH or JM rom (QDOS 1.03 or lower) and if so, return 128
* the only value supported by these rom versions (fixed line width of
* 128 bytes wide only)
* need to preserve channel ID in a0, which is smashed by mt.inf
        move.l  a0,d6           ;save channel id which was in a0
        moveq   #0,d0           ;call mt.inf to get system information
        trap    #1              ;call mt.inf
        move.l  d6,a0           ;return channel ID to a0
* d2.l now contains the ASCII version number (n.nn)
* remove the decimal point to make things easier
* as some national ROM versions contain a letter code rather than "."
        move.l  #128,d0         ;in case it's an AH/JM rom...
        and.l   #$FF00FFFF,d2   ;remove '.'
        cmpi.l  #$31003034,d2   ;>=1.04? <"1"><nul><"0"><"4">
        bcs     retfp1          ;no, return 128

* prepare to fetch info from definition block
        moveq   #9,d0           ;prepare to use the channel
        moveq   #$64,d1         ;offset in chan def block for sd.linel
        moveq   #-1,d3          ;infinite timeout
        lea.l   extop_w,a2      ;the routine to be called by sd.extop
        trap    #3              ;invoke sd.extop
* d1 now contains the value required (screen base)
        ext.l   d1              ;stretch to long word
        move.l  d1,d0           ;put required value in d0 for retfp
        bra     retfp1          ;and return it
        
* LET mde = DMODE
* no parameters in the basic call; return 0 or 8 for mode number
dmode   move.l  a5,d0           ;pointer to last parameter
        sub.l   a3,d0           ;subtract pointer to parameter 1
        lsr.l   #3,d0           ;bytes DIV 8 = number of parameters
        bne     badparam        ;oops, silly boy
        moveq   #$10,d0         ;mt.dmode
        move.b  #-1,d1          ;READ mode
        move.b  #-1,d2          ;READ display (incidental here)
        trap    #1
* d1.b=display mode number
        moveq.l #6,d7           ;6 bytes needed on stack to retn float
        clr.l   d0              ;empty top 3 bytes
        move.b  d1,d0           ;mode number for conversion to fp
        bra     retfp           ;send it back to basic

* LET sys_variables = SYS_VAR
sys_var move.l  a5,d0           ;pointer to last parameter
        sub.l   a3,d0           ;subtract pointer to 1st parameter
        lsr.l   #3,d0           ;number of parameters (8 bytes/param)0
        bne     badparam        ;oops
        moveq.l #0,d0           ;mt.inf
        trap    #1              ;call mt.inf
* returns d2.l = QDOS ASCII version number n.nn
* and a0 = pointer to system variables
        move.l  a0,d0           ;system variables to d0
        moveq.l #6,d7           ;space needed
        bra     retfp           ;return it to basic

* the iop.flim based calls follow
* LET w=FLIM_W(#channel)        LET h=FLIM_H(#channel)
* LET x=FLIM_X(#channel)        LET y=FLIM_Y(#channel)
* routines need a 4 byte storage area required by iop.flim for
* width,height,x,y
store   ds.w    4
flim_w  move.w  #512,d5         ;default width
        moveq.l #0,d6           ;offset of width value from "store"
        bra.s   flim            ;branch to common part of routine

flim_h  move.w  #256,d5         ;default height
        moveq.l #2,d6           ;offset of height value from "store"
        bra.s   flim            ;branch to common part of routine

flim_x  move.w  #0,d5           ;default x
        moveq.l #4,d6           ;offset of x value from "store"
        bra.s   flim            ;branch to common part of routine
        
flim_y  move.w  #0,d5           ;default y
        moveq.l #6,d6           ;offset of y value from "store"

* common routine, with pre-assigned values in d5+d6
flim    move.w  $118,a2         ;ca.gtlin
        jsr     (a2)            ;fetch 1 parameter
        bne.s   return          ;oops!
        subq.w  #1,d3           ;1 parameter
        bne.s   badparam        ;oops!
* we now have the channel number on the stack
        moveq.l #2,d7           ;2 more bytes will be needed
        move.l  0(a6,a1.l),d0   ;get channel number to d0
        bmi     retfp1          ;if negative, return as function value
        bsr     chan2id         ;channel # to ID in a0
        bmi     retfp1          ;oops,return the -ve number as fn value

        moveq   #-1,d3          ;timeout
        clr.l   d2              ;has to be zero
        move.l  a1,$58(a6)      ;store RI stack pointer
        lea.l   store,a1        ;point to 8 byte area
        moveq   #$6C,d0         ;iop.flim
        trap    #3
        move.l  $58(a6),a1      ;recover RI stack pointer
        tst.w   d0              ;was there any error?
        beq.s   noerr           ;no, fetch value from store to return
* since iop.flim appears not to exist, we'll assume a 512x256 display
        move.w  d5,d0           ;the default value for this function
        bra.s   extnd           ;and return that default value
* fetch values for return
noerr   lea.l   store,a2        ;point to four word store area
        move.w  0(a2,d6.l),d0   ;fetch value
extnd   ext.l   d0              ;convert to long word
        bra     retfp1          ;return as floating point value
* utility routines shared by functions
badparam moveq  #-15,d0         ;err.bp
return  rts                     ;go back to basic with d0 error report
notopen moveq   #-6,d0          ;err.no
        rts

* LET pe = PTR_ENV(#channel)
ptr_env move.w  $118,a2         ;ca.gtlin
        jsr     (a2)            ;fetch 1 parameter
        bne     return          ;oops
        subq.w  #1,d3           ;1 parameter?
        bne     badparam        ;oops
* we have channel number on stack
        moveq.l #2,d7           ;2 extra bytes will be needed
        move.l  0(a6,a1.l),d0   ;channel number
        bmi     retfp1          ;if negative, return as function value
        bsr     chan2id         ;convert channel# in d0 to ID in a0
        bmi     retfp1          ;if -ve return as fn value
        moveq.l #-1,d3          ;infinite timeout
        move.l  a1,$58(a6)      ;store RI stack pointer
        moveq.l #$70,d0         ;iop.pinf (Get Pointer Info) trap
        trap    #3
        move.l  $58(a6),a1      ;recover RI stack pointer
        tst.l   d0              ;is there a pointer interface?
        bne.s   no              ;no there isn't
yes     moveq.l #1,d0           ;1=yes, present
        bra     retfp1          ;return 1 as function value
no      moveq.l #0,d0           ;0=no, not present
        bra     retfp1          ;return 0 as function value

* LET wm = WIN_MAN(#channel)
win_man move.w  $118,a2         ;ca.gtlin
        jsr     (a2)            ;fetch 1 parameter
        bne     return          ;oops
        subq.w  #1,d3           ;1 parameter?
        bne     badparam        ;oops
* we have channel number on stack
        moveq.l #2,d7           ;2 extra bytes will be needed
        move.l  0(a6,a1.l),d0   ;channel number
        bmi     retfp1          ;if negative, return as function value
        bsr     chan2id         ;convert channel# in d0 to ID in a0
        bmi     retfp1          ;if -ve return as fn value
        moveq.l #-1,d3          ;infinite timeout
        move.l  a1,$58(a6)      ;store RI stack pointer
        moveq.l #$70,d0         ;iop.pinf (Get Pointer Info) trap
        trap    #3
        tst.l   d0              ;is there pointer interface?
        beq.s   ptryes          ;yes there is
* there's no pointer interface, so there can't be a window manager test!
        move.l  $58(a6),a1      ;recover RI stack pointer
        bra.s   no              ;no pointer interface, can't be window manager
ptryes  move.l  a1,d0           ;window manager vector in a1
        move.l  $58(a6),a1      ;restore RI stack pointer
        tst.l   d0              ;if vector=0, no wman
        ble.s   no              ;no there isn't
        bra.s   yes             ;yes there is

* LET v$ = PTRVER$(#channel) returns pointer version n.nn
ptr_ver move.w  $118,a2         ;ca.gtlin
        jsr     (a2)            ;fetch 1 parameter
        bne     return          ;oops
        subq.w  #1,d3           ;1 parameter?
        bne     badparam        ;oops
* we have channel number on stack, so can return values
        move.l  0(a6,a1.l),d0   ;channel number
        bmi.s   retnul          ;if negative, return "" as function value
        bsr     chan2id         ;convert channel# in d0 to ID in a0
        bmi.s   retnul          ;if -ve return "" as function value
        move.l  a1,$58(a6)      ;store RI stack pointer in bv.rip
* we are now going to call iop.pinf to check if PE available
        moveq.l #-1,d3          ;infinite timeout
        moveq.l #$70,d0         ;iop.pinf (Get Pointer Info) trap
        trap    #3              ;returns version number in d1.l

        move.l  $58(a6),a1      ;recover RI stack pointer
        tst.l   d0              ;is there a pointer interface?
        bne.s   retnul          ;oops, no there isn't

* need 2 extra bytes on RI stack for n.nn string return
retvstr move.l  d1,-(a7)        ;stack version number temporarily
        move.l  a1,$58(a6)      ;store RI stack pointer
        moveq.l #2,d1           ;2 extra bytes needed on RI stack
        move.w  $11A,a0         ;bv.chrix vector
        jsr     (a0)            ;get 2 extra bytes
        move.l  (a7)+,d1        ;recover version number
        move.l  $58(a6),a1      ;recover RI stack pointer
        subq.l  #2,a1           ;point to length word
        move.l  a1,$58(a6)      ;store RI stack pointer in bv.rip
        move.w  #4,0(a6,a1.l)   ;4 digit string to return
        move.l  d1,2(a6,a1.l)   ;n.nn pointer version number
retstr  moveq   #1,d4           ;string result
        moveq   #0,d0           ;no error
        rts                     ;back to basic
retnul  move.l  $58(a6),a1      ;recover RI stack pointer
        addq.l  #2,a1           ;nul string needs 2 bytes, not 4
        move.l  a1,$58(a6)      ;store in bv.rip
        clr.w   0(a6,a1.l)      ;string length 0
        bra.s   retstr

* LET v$ = OS_VER$   returns operating systm version number
os      moveq   #6,d1           ;6 bytes needed to return version$
        move.w  $11A,a0         ;bv.chrix vector
        jsr     (a0)            ;get the 6 extra bytes
        move.l  $58(a6),a1      ;recover RI stack pointer
        subq.l  #6,a1           ;point to string length word
        moveq   #0,d0           ;mt.inf
        trap    #1              ;get system info
        move.w  #4,0(a6,a1.l)   ;4 character string to return
        move.l  d2,2(a6,a1.l)   ;QDOS version number in d2.l
        moveq   #1,d4           ;string result
        move.l  a1,$58(a6)      ;store RI stack pointer
        moveq   #0,d0           ;no error
        rts

* LET gd = GD2(#channel)  returns 1 if GD2 present, 0 if not
gd2     move.w  $118,a2         ;ca.gtlin
        jsr     (a2)            ;fetch 1 parameter
        bne     return          ;oops
        subq.w  #1,d3           ;1 parameter?
        bne     badparam        ;oops
* we have channel number on stack
        moveq.l #2,d7           ;2 extra bytes will be needed
        move.l  0(a6,a1.l),d0   ;channel number
        bmi     retfp1          ;if negative, return as function value
        bsr     chan2id         ;convert channel# in d0 to ID in a0
        bmi.s   retfp1          ;if -ve return as fn value
        moveq.l #-1,d3          ;infinite timeout
        move.l  a1,$58(a6)      ;store RI stack pointer
        moveq.l #$70,d0         ;iop.pinf (Get Pointer Info) trap
        trap    #3
        move.l  $58(a6),a1      ;recover RI stack pointer
        tst.l   d0              ;is there a pointer interface?
        bne.s   nogd2           ;no there isn't, return 0
        clr.w   d1              ;don't set anything, only get
        clr.w   d2              ;address of PE linkage block
        move.w  #-1,d3          ;infinite timeout
        moveq   #$6F,d0         ;iop.slnk
        trap    #3              ;call iop.slnk
        moveq   #1,d0           ;assume GD2 present
        move.l  $128(a1),d1     ;check GD2 flag location
        move.l  $58(a6),a1      ;recover RI stack pointer again
        cmp.l   #"PTR2",d1      ;"PTR2"=value for GD2
        beq.s   retfp1          ;yes, GD2 present, return 1
nogd2   moveq   #0,d0           ;no GD2 present
        bra     retfp1          ;return 0

* LET v$ = WMAVER$(#channel) returns wman version n.nn
wma_ver move.w  $118,a2         ;ca.gtlin
        jsr     (a2)            ;fetch 1 parameter
        bne     return          ;oops
        subq.w  #1,d3           ;1 parameter?
        bne     badparam        ;oops
* we have channel number on stack, so can return values
        move.l  0(a6,a1.l),d0   ;channel number
        bmi     retnul          ;if negative, return "" as function value
        bsr     chan2id         ;convert channel# in d0 to ID in a0
        bmi     retnul          ;if -ve return "" as function value
        move.l  a1,$58(a6)      ;store RI stack pointer in bv.rip
* we are now going to call iop.pinf to check if PE available
        moveq.l #-1,d3          ;infinite timeout
        moveq.l #$70,d0         ;iop.pinf (Get Pointer Info) trap
        trap    #3              ;returns version number in d1.l

        tst.l   d0              ;is there a pointer interface?
        bne     retnul          ;oops, no there isn't
        cmpa.l  #0,a1           ;is there a window manager vector?
        beq     retnul          ;oops, no there isn't
        move.l  0(a1),d1        ;get long word version number at wman vector
        move.l  $58(a6),a1      ;recover RI stack pointer
        bra     retvstr         ;return the version number in d1.L


* convert and return long word in d0 into a floating point number and
* return (used with permission of Simon N. Goodwin).
* d7.w contains no. of extra bytes required on RI stack over and above
* what's on the RI stack parameters
retfp1  move.l  a1,$58(a6)      ;store RI stack pointer if relevant
retfp   move.w  d0,d1           ;d1 will be exponent
        move.l  d0,d2           ;d2 will be mantissa
        beq.s   normalised      ;zero is a trivial case
        move.w  #2079,d1        ;first guess at exponent
        add.l   d0,d0           ;already normalised?
        bvs.s   normalised
        subq.w  #1,d1           ;no, halve exponent weight
        move.l  d0,d2           ;double mantissa to match
        moveq   #16,d3          ;try a 16 bit shift
normalise move.l d2,d0          ;take copy of mantissa
        asl.l   d3,d0           ;shift mantissa d3 places
        bvs.s   too_far         ;overflow - must shift less
        sub.w   d3,d1           ;correct exponent for shift
        move.l  d0,d2           ;new mantissa is more normal
too_far asr.w   #1,d3           ;halve shift distance
        bne.s   normalise       ;try shift of 8,4,2 and 1
* is extra space required on RI stack? d7 holds number of bytes
* needed over and above existing parameter space
* need to store anything important not preserved by bv.chrix call
normalised movem.l d1-d2,-(a7)  ;store exponent/mantissa
        tst.l   d7              ;if non-zero, create more space
        beq.s   enough          ;no more needed
        move.l  d7,d1           ;no. of extra bytes needed
        move.w  $11A,a0         ;bv.chrix vector
        jsr     (a0)            ;do it
        move.l  $58(a6),a1      ;recover ri pointer
        sub.w   d7,a1           ;adjust by no. of bytes
        move.l  a1,$58(a6)      ;additional bytes pointed to now
enough  movem.l (a7)+,d1-d2     ;recover exponent/mantissa
        move.l  d2,2(a1,a6.l)   ;stack mantissa
        move.w  d1,0(a1,a6.l)   ;stack exponent
        moveq   #2,d4           ;signal FP result
        moveq   #0,d0           ;no errors PLEASE!
        rts

* subroutine to convert channel # in d0 to 4-byte channel ID in a0
chan2id tst.l   d0              ;can't be negative
        bmi.s   retfp1          ;oops, was negative...
        mulu    #$28,d0         ;convert into channel table offset
        add.l   $30(a6),d0      ;add start of table
        cmp.l   $34(a6),d0      ;past end of table?
        bge     notopen         ;yes, channel not open
        movea.l 0(a6,d0.l),a0   ;no-get ID into a0.l
        tst.l   0(a6,d0.l)      ;is channel open?
        bmi     notopen         ;no, return err.no
        moveq   #0,d0           ;yes, return OK code
        rts                     ;and exit
